/*
 * Copyright (c) 2004-2006 Erez Zadok,
 * Copyright (c) 2004-2006 Nikolai Joukov,
 * Copyright (c) 2006      Chaitanya Patti,
 * Copyright (c) 2004      Abhishek Rai,
 * Copyright (c) 2004-2006 Stony Brook University
 */


/* geneic loadbal.h file: generates data for user level analyzers */
#ifndef __LOADBAL_H_
#define __LOADBAL_H_
#ifdef __KERNEL__
#ifdef RAIF_LB

/* for kernel thread support */
#include <linux/init.h>
#include <linux/sched.h>

/* mischellaneous macros */
#define sbload(sb, index)	(stopd(sb)->load[index])
#define sbscratchpool(sb)	(stopd(sb)->scratchpool)
#define sbrandbuf(sb)	(stopd(sb)->randbuf)
#define sbpool(sb)	(stopd(sb)->pool)
#define sbpoolend(sb)	(stopd(sb)->poolend)
#define sbpoolcur(sb)	(stopd(sb)->poolcur)

#define sbpoolsize(sb)	(4 + (4 * MAX_RATIO *  (sbnum(sb)>0 ? sbnum(sb)-1 : 0) ))


/* used by the latency estimation mechanism */
#define ALPHA	16
#define BETA	4

/* ration of maximum load to minimum load among the branches, an artificial restrxn imposed for convenience */
#define MAX_RATIO 50

/* macros for reading latency measurements */
#define READ_LAT(SB, index)   (atomic_read(& (stopd(SB)->lat[index].read.avg) ))
#define WRITE_LAT(SB, index)  (atomic_read(& (stopd(SB)->lat[index].write.avg) ))
#define LOOKUP_LAT(SB, index) (atomic_read(& (stopd(SB)->lat[index].lookup.avg) ))

#define LBFS_STR_read	"r"
#define LBFS_STR_write	"w"
#define LBFS_STR_lookup	"l"

#define LBFS_READ_WT	1
#define LBFS_WRITE_WT	2
#define LBFS_LOOKUP_WT	1


extern int load_update(void *);
extern void init_pool(struct super_block *);
extern void recalc_light(struct super_block *);
extern void update_lat(struct timeval *,
		oplat_t *, struct timeval *, struct timeval *, char *);


static inline unsigned int lb_getload(super_block_t *sb, int index) {
    return 	( READ_LAT(sb, index)*LBFS_READ_WT 
    		+ WRITE_LAT(sb, index)*LBFS_WRITE_WT
		+ LOOKUP_LAT(sb, index)*LBFS_LOOKUP_WT
		);
}

/* LBFS_ALPHA should be set to as many micro-seconds after which, u think, 
 * the new observed lat and old lat have equal weightage. 
 * Probably doesn't make sense to have smaller than common 
 * inter-operation (readpage/writepage) time. In fact, it should be somewhat larger than
 * inter-operation time, to allow for some hysteris, and stability.
 * 
 * Inter-operation times depend hugely on the type of operation. While consecutive
 * writes have inter-writepage time of >1ms usually, it is smaller for reads.
 * But we can worry about it later, as not sure if writepage does actually go to disk etc
 */
/* following values were set for vmware read/write/lookup */
//#define LBFS_ALPHA_read		1024
//#define	LBFS_ALPHA_write	128
//#define LBFS_ALPHA_lookup	16

static inline int lb_fastest_branch(struct inode *inode) {
	int index;
	struct super_block *sb = inode->i_sb;

#ifndef LB_ALTERNATE
	index = sbpool(sb)[sbpoolcur(sb)];
	if (++sbpoolcur(sb) > sbpoolend(sb))
		sbpoolcur(sb) = 0;
	return index;
#else
	index = stopd(sb)->curr;
	if (++stopd(sb)->curr > sbnum(sb))
		stopd(sb)->curr = 0;
	return index;
#endif
}


#define LB_UPDATE_LAT(SB, index, op, b4, af)	spin_lock( & stopd(SB)->lat_lock[index]); \
					update_lat( &stopd(SB)->lat[index].last, &stopd(SB)->lat[index].op, &b4, &af, LBFS_STR_##op); \
					spin_unlock( & stopd(SB)->lat_lock[index]);


#define LB_UPDATE_PREPARE \
	do_gettimeofday(&b4);

#define LB_UPDATE_COMMIT_I(ino, bindex, op) \
	do_gettimeofday(&af);\
	LB_UPDATE_LAT((ino)->i_sb, bindex, op, b4, af);

#define LB_UPDATE_COMMIT_F(fil, bindex, op) \
	do_gettimeofday(&af);\
	LB_UPDATE_LAT((fil)->f_vfsmnt->mnt_sb, bindex, op, b4, af);

#else

#define LB_UPDATE_PREPARE
#define LB_UPDATE_COMMIT_I(ino, bindex, op)
#define LB_UPDATE_COMMIT_F(fil, bindex, op)

#endif /* RAIF_LB */
#endif /* __KERNEL__ */
#endif	/* not __LOADBAL_H_ */
