/*
 * Copyright (c) 1997-2006 Erez Zadok
 * Copyright (c) 1997-2006 Nikolai Joukov
 * Copyright (c) 2001-2006 Stony Brook University
 * Copyright (c) 1997-2000 Columbia University
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 *  $Id: raif.h,v 1.81 2006/12/13 17:00:21 kolya Exp $
 */

#ifndef __RAIF_H_
#define __RAIF_H_

#ifdef __KERNEL__


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/spinlock.h>
#define wq_write_lock_irqsave(lock,flags) spin_lock_irqsave(lock,flags)
#define wq_write_unlock(lock) spin_unlock(lock)
#define wq_write_lock_irq(lock) spin_lock_irq(lock)
#define wq_write_unlock_irqrestore(lock,flags) spin_unlock_irqrestore(lock,flags)
#endif

#include <linux/list.h>

/* fist file systems superblock magic */
#define RAIF_SUPER_MAGIC 0xf15f

/* raif root inode number */
#define RAIF_ROOT_INO	 1

/* used in readdir/filldir combo in file.c for duplicate elimination */
#define HASHTABLE_SIZE  13

#define raif_interpose(x, y, z)		raif_interpose_debug((x), (y), (z), __FILE__, __LINE__, __FUNCTION__)
extern int raif_interpose_debug(dentry_t *dentry, super_block_t *sb, int flag, char *file, int line, const char *function);

/* raif inode data in memory */
struct raif_inode_info {
	inode_t ** uii_inode;
	int b_num;
#ifdef RAIF_TAG
	int level;
	int start_branch;
	int striping_unit_shift;
#endif
#if defined(RAIF_LB) && defined(LB_AFFINITY)
	int lb_aff;
#endif 
	/*Stuff for readdir */
	spinlock_t uii_rdlock;
	struct list_head uii_readdircache;
	int uii_rdcount;
	int uii_hashsize;
	int uii_cookie;
};

#ifdef RAIF_TAG
	struct rule {
		struct rule* next;
		int b_num;
		int level;
		int start_branch;
		unsigned int striping_unit_shift;
		unsigned int mask;
		unsigned int special; 
		/*
		* special flag:
		*
		* for local rules: it indicates if this is special rule introduced during rename. it affects only a specific file (the target of the rename)
		*/
		char regexp[0];
	};
#endif

/* raif dentry data in memory */
struct raif_dentry_info {
	dentry_t ** udi_dentry;
	int b_num;
#ifdef RAIF_TAG
	unsigned int mask;
	//the following must strictly have been a union
	//union {
	struct rulelookup {
		struct lookup {
			int level;
			int start_branch;
			unsigned int striping_unit_shift;
		} lookup;
		struct rules {
			struct rule* rule; // singly-linked list
			struct rule* local_rule; // rules for this directory only
			rwlock_t spin; // to protect the list of rules
		} rules;
	} d;
#endif
};

#ifdef RAIF_LB
/* lbfs latency measurement */
typedef struct {
	 atomic_t avg;
	 int dev;
} oplat_t;
typedef struct {
	 oplat_t read, write, lookup;
	 struct timeval last;
} latency_t;
#endif /* RAIF_LB */

/* raif super-block data in memory */
struct raif_sb_info {
	super_block_t **usi_sb;
	vfs_mount_t **usi_hidden_mnt;
	int b_num;
	int level;
	int striping_unit_shift;


#ifdef RAIF_LB
	int *load;
	char *pool, *scratchpool;
	char *randbuf;
	/* poolend is the index of the last entry in the pool */
	int poolend;
	int poolcur;
	atomic_t die;
	wait_queue_head_t clean;
	/* later modify latency_t to have fields for all of read/write/lookup */
	latency_t *lat;
	spinlock_t *lat_lock;
#ifdef LB_ALTERNATE
	int curr;
#endif /* LB_ALTERNATE*/
#endif /* RAIF_LB */

	int is_lower_nfs;

};

/* file private data. */
struct raif_file_info {
	struct file ** ufi_file;
	int b_num;
/*	int ufi_bindex;
	struct list_head dir_list[HASHTABLE_SIZE];*/
	struct raif_dir_state *rdstate;
};

/*
 * MACROS:
 */
// File TO Private Data
#define ftopd(file) ((struct raif_file_info *)((file)->private_data))
// File to Hidden File based on index
#define ftohf_index(file, index) ((ftopd(file))->ufi_file[index])
// File to Hidden File Pointers
#define ftohf_ptr(file)  (ftopd(file)->ufi_file)
// Macros to refer to branch indices
#define fbnum(file) (ftopd(file)->b_num)
// Initialize private data for file
#define init_ftopd(file) memset(ftopd(file), 0, sizeof(struct raif_file_info));
/*
#define init_ftopd(file) \
	 {\
		int i = 0; \
		memset(ftopd(file), 0, sizeof(struct raif_file_info));\
		ftopd(file)->ufi_bindex = -1; \
		for (i = 0; i < HASHTABLE_SIZE; i++) \
			INIT_LIST_HEAD(&(ftopd(file)->dir_list[i])); \
	 }
*/
// Initialize array pointers for hidden files
#define init_ftohf_ptr(file, num) memset(ftohf_ptr(file), 0, (sizeof(file_t *) * num))

// Inode TO Private Data
#define itopd(ino) ((struct raif_inode_info *)(ino)->u.generic_ip)
// Inode To Hidden Inode based on index
#define itohi_index(ino, index) (itopd(ino)->uii_inode[index])
// Inode To Hidden Inode Pointers 
#define itohi_ptr(ino)  (itopd(ino)->uii_inode)
// Macros to refer to branch indices
#define ibnum(ino) (itopd(ino)->b_num)

#ifdef RAIF_TAG
#define START_BRANCH(ino) (itopd(ino)->start_branch)
#define STRIPING_UNIT_SHIFT(ino) (itopd(ino)->striping_unit_shift)
#define LEVEL(ino) (itopd(ino)->level)
#else
//#define AUTH_BRANCH(ino) 0
#define START_BRANCH(ino) 0
#define STRIPING_UNIT_SHIFT(ino) 0
#define LEVEL(ino) 0
#endif
#define STRIPING_UNIT_SIZE(ino) (1 << STRIPING_UNIT_SHIFT(ino))

/*Maximum striping unit shift*/
#define RAIF_MAX_SUSH 64

// Initialize private data for inode
#if defined(RAIF_LB) && defined(LB_AFFINITY)
#define init_itopd(ino)  \
	{\
		memset(itopd(ino), 0, sizeof(struct raif_inode_info));\
		itopd(ino)->lb_aff = -1;\
		/*RAIF_BRANCH_MNGMNT atomic_set(&itopd(ino)->uii_generation, atomic_read(&stopd(ino->i_sb)->usi_generation));*/\
	}
#else
#define init_itopd(ino)  \
	{\
		memset(itopd(ino), 0, sizeof(struct raif_inode_info));\
		/*RAIF_BRANCH_MNGMNT atomic_set(&itopd(ino)->uii_generation, atomic_read(&stopd(ino->i_sb)->usi_generation));*/ \
	}
#endif
// Initialize array pointers for inode
#define init_itohi_ptr(inode, num) memset(itohi_ptr(inode), 0, (sizeof(inode_t *) * num))

// Superblock TO Private Data
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#define stopd(super) ((struct raif_sb_info *)(super)->u.generic_sbp)
#define stopd_lhs(super) ((super)->u.generic_sbp)
#else
#define stopd(super) ((struct raif_sb_info *)(super)->s_fs_info)
#define stopd_lhs(super) ((super)->s_fs_info)
#endif

// Superblock to Hidden Superblock based on branch index
#define stohs_index(super, global_bindex) (stopd(super)->usi_sb[global_bindex])
// Superblock to Hidden Array of super blocks
#define stohs_ptr(super) (stopd(super)->usi_sb)
// Superblock to Hidden Mount point
#define stohm_index(super, global_bindex) (stopd(super)->usi_hidden_mnt[global_bindex])
// Superblock to Hidden Mount Pointer
#define stohiddenmnt_ptr(super) (stopd(super)->usi_hidden_mnt)
// Macros to refer to the branch indices
#define sbnum(sb) stopd(sb)->b_num
// Initialize private data for super block
#define init_stopd(sb) \
	{\
		memset(stopd(sb), 0, sizeof(struct raif_sb_info)); \
		stopd(sb)->level = -1; \
		stopd(sb)->striping_unit_shift = -1; \
		stopd(sb)->is_lower_nfs = 0; \
		/*RAIF_BRANCH_MNGMNT atomic_set(&stopd(sb)->usi_generation, 1);*/\
	}
// Initialize array pointers for hidden superblock
#define init_stohs_ptr(sb, num) memset(stohs_ptr(sb), 0, (sizeof(super_block_t *) * num))
#define init_stohiddenmnt_ptr(sb, num) memset(stohiddenmnt_ptr(sb), 0, (sizeof(vfs_mount_t *) * num))

// Superblock to unusable branches value based on branch index
//#define unusable_branch_index(super, index) (stopd(super)->unusable_branches[index])


// Dentry TO Hidden Dentry
#define dtopd(dentry) ((struct raif_dentry_info *)(dentry)->d_fsdata)
// Dentry to Hidden Dentry based on index
#define dtohd_index(dent, index) (dtopd(dent)->udi_dentry[index])
// Dentry to Hidden Dentry Pointer
#define dtohd_ptr(dentry) (dtopd(dentry)->udi_dentry)


// Macros to refer to branch indices
#define dbnum(dentry) (dtopd(dentry)->b_num)
// Initialize private data for inode
#define init_dtopd(dentry, sb) \
	{\
		memset(dtopd(dentry), 0, sizeof(struct raif_dentry_info));\
		dtopd(dentry)->b_num = stopd(sb)->b_num;\
		/* RAIF_BRANCH_MNGMNT atomic_set(&dtopd(dentry)->udi_generation, atomic_read(&stopd(sb)->usi_generation));*/\
	}
// Initialize raif dentry info
#define init_dentry_info(dentry) \
	{\
		memset(dentry, 0, sizeof(struct raif_dentry_info));\
		dentry->b_num = 0;\
	}
// Initialize array pointers for sb
#define init_dtohd_ptr(dentry, num) memset(dtohd_ptr(dentry), 0, (sizeof(dentry_t *) * num))

//#define sbt(sb) ((sb)->s_type->name)

/* macros to simplify non-SCA code */
#define CURRENT_HIDDEN_PAGE hidden_page
#define CURRENT_HIDDEN_PAGEDATA hidden_page_data
#define CURRENT_HIDDEN_PAGEINDEX page->index

/*
 * EXTERNALS:
 */
extern struct file_operations raif_main_fops;
extern struct file_operations raif_dir_fops;
extern struct inode_operations raif_main_iops;
extern struct inode_operations raif_dir_iops;
extern struct inode_operations raif_symlink_iops;
extern struct super_operations raif_sops;
extern struct dentry_operations raif_dops;
extern struct vm_operations_struct raif_shared_vmops;
extern struct vm_operations_struct raif_private_vmops;
extern struct address_space_operations raif_aops;

//extern page_t *raif_get1page(file_t *file, int index);
//extern int raif_fill_zeros(file_t *file, page_t *page, unsigned from);

/* d_release function. */
void raif_d_release(dentry_t *dentry);

/* replicates the directory structure upto given dentry in given branch */
extern dentry_t *raif_create_parent_dir(inode_t *dir, dentry_t *dentry, int bindex);

/* partial lookup */
extern int raif_partial_lookup(dentry_t *dentry);

extern int create_dir_whs(dentry_t *dentry, int cur_index) ;

/* Is this directory empty: 0 if it is empty, -ENOTEMPTY if not. */
int check_empty(dentry_t *dentry);

/* The values for raif_interpose's flag. */
#define INTERPOSE_DEFAULT	0
#define INTERPOSE_LOOKUP	1
#define INTERPOSE_REVAL		2


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
/* Extended attribute functions. */
int raif_getxattr(struct dentry *dentry, const char *name, void *value, size_t size);
int raif_removexattr(struct dentry *dentry, const char *name);
int raif_listxattr(struct dentry *dentry, char *list, size_t size);

#ifndef FIST_SETXATTR_CONSTVOID
int raif_setxattr(struct dentry *dentry, const char *name, void *value, size_t size, int flags);
#else /* not FIST_SETXATTR_CONSTVOID */
int raif_setxattr(struct dentry *dentry, const char *name, const void *value, size_t size, int flags);
#endif /* FIST_SETXATTR_CONSTVOID */

#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20) */

#define copy_inode_size(dst, src) \
	dst->i_size = src->i_size; \
	dst->i_blocks = src->i_blocks; \
	fist_dprint(7, "copy_inode_size: i_size=%d, i_blocks=%d\n", src->i_size, src->i_blocks);


/* returns the sum of the n_link values of all the underlying inodes of the passed inode */
static inline int
get_nlinks(inode_t *inode) 
{
	int sum_nlinks = 1, bindex;
	inode_t *hidden_inode;

	for (bindex = 0; bindex < ibnum(inode); bindex++) {
		hidden_inode = itohi_index(inode, bindex);
	
		if (!hidden_inode)
			continue;
		sum_nlinks += (hidden_inode->i_nlink - 1);
	}
	return sum_nlinks;
}


static inline void
fist_copy_attr_atime(inode_t *dest, const inode_t *src)
{
	PASSERT(dest);
	PASSERT(src);
	dest->i_atime = src->i_atime;
}
static inline void
fist_copy_attr_times(inode_t *dest, const inode_t *src)
{
	PASSERT(dest);
	PASSERT(src);
	dest->i_atime = src->i_atime;
	dest->i_mtime = src->i_mtime;
	dest->i_ctime = src->i_ctime;
}
static inline void
fist_copy_attr_timesizes(inode_t *dest, const inode_t *src)
{
	fist_copy_attr_times(dest, src);
	copy_inode_size(dest, src);
}
static inline void
fist_copy_attr_allbutsizes(inode_t *dest, const inode_t *src)
{
	//print_entry_location();
	PASSERT(dest);
	PASSERT(src);
	dest->i_mode = src->i_mode;
	dest->i_nlink = get_nlinks(dest);
	dest->i_uid = src->i_uid;
	dest->i_gid = src->i_gid;
	dest->i_rdev = src->i_rdev;
	fist_copy_attr_times(dest, src);
	dest->i_blksize = src->i_blksize;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,12)
	dest->i_blkbits = src->i_blkbits;
#endif /* linux 2.4.12 and newer */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	dest->i_attr_flags = src->i_attr_flags;
#else
	dest->i_flags = src->i_flags;
#endif
}


static inline void
fist_copy_attr_all(inode_t *dest, const inode_t *src)
{
	fist_copy_attr_allbutsizes(dest, src);
	copy_inode_size(dest, src);
}


/* Copied from fs/dcache.c. */
/*
 * Release the dentry's inode, using the filesystem
 * d_iput() operation if defined.
 * Called with dcache_lock held, drops it.
 */
static inline void dentry_iput(struct dentry * dentry)
{
	struct inode *inode = dentry->d_inode;
	if (inode) {
		dentry->d_inode = NULL;
		list_del_init(&dentry->d_alias);
		spin_unlock(&dcache_lock);
		if (dentry->d_op && dentry->d_op->d_iput)
			dentry->d_op->d_iput(dentry, inode);
		else
			iput(inode);
	} else
		spin_unlock(&dcache_lock);
}

dentry_t *raif_lookup_backend(inode_t *dir, dentry_t *dentry, int interpose_flag);
int is_stale_inode(struct inode * inode);
void make_stale_inode(struct inode * inode);

#ifdef RAIF_TAG
#define RAIF_RULES_FILE_NAME		".raif.rules"
#define RAIF_RULES_FILE_NAME_LEN	(sizeof(RAIF_RULES_FILE_NAME) - 1)
#define RAIF_TMPREFIX			".raif9421temp."
#define RAIF_TMPREFIX_LEN		(sizeof(RAIF_TMPREFIX) - 1)
#define RAIF_MAX_RULE_SIZE 200

int raif_rule_interpose(dentry_t* new_dentry);
int raif_rule_lookup(dentry_t* new_dentry);
int raif_rule_rename(dentry_t* old_dentry, dentry_t* new_dentry);
int raif_rule_global_bindex(dentry_t* dentry, int bindex);
int raif_rule_unlink_file(dentry_t *);
void raif_rule_free(struct rule *);
int raif_rule_remove_special(dentry_t * dentry);
#define GLOBAL_BINDEX(d,b) raif_rule_global_bindex((d),(b))
#else
#define GLOBAL_BINDEX(d,b) (b)
#endif

/*
 * structure for making the linked list of entries by readdir on left branch
 * to compare with entries on right branch
 */
struct filldir_node {
	struct list_head file_list;	// list for directory entries
	char *name;		// name entry
	int hash;		// name hash
	int namelen;		// name len since name is not 0 terminated
	int bindex;
	char iname[DNAME_INLINE_LEN_MIN];	// Inline name, so we don't need to separately kmalloc small ones
};

/* Cache creation/deletion routines. */
void destroy_filldir_cache(void);
int init_filldir_cache(void);
int init_inode_cache(void);
void destroy_inode_cache(void);
int init_dentry_cache(void);
void destroy_dentry_cache(void);

/* Initialize and free readdir-specific  state. */
int GetNextBranch(int minSet, int CurrIndex);
int init_rdstate(struct file *file);
struct raif_dir_state *alloc_rdstate(struct inode *inode, int bindex);
struct raif_dir_state *find_rdstate(struct inode *inode, loff_t fpos);
void free_rdstate(struct raif_dir_state *state);
int add_filldir_node(struct raif_dir_state *rdstate, const char *name,
		     int namelen, int bindex);
struct filldir_node *find_filldir_node(struct raif_dir_state *rdstate,
				       const char *name, int namelen);

/* Directory hash table. */
struct raif_dir_state {
	unsigned int uds_cookie;	/* The cookie, which is based off of uii_rdversion */
	unsigned int uds_offset;	/* The entry we have returned. */
	int uds_bindex;
	loff_t uds_dirpos;	/* The offset within the lower level directory. */
	int uds_size;		/* How big is the hash table? */
	int uds_hashentries;	/* How many entries have been inserted? */
	int minSet;				/* The minimal set of branches to be read */
	unsigned long uds_access;
	/* This cache list is used when the inode keeps us around. */
	struct list_head uds_cache;
	struct list_head uds_list[0];
};
/* We can only use 32-bits of offset for rdstate --- blech! */
#define DIREOF (0xfffff)
#define RDOFFBITS 20		/* This is the number of bits in DIREOF. */
#define MAXRDCOOKIE (0xfff)
/* Turn an rdstate into an offset. */
static inline off_t rdstate2offset(struct raif_dir_state *buf)
{
	off_t tmp;
	tmp =
	    ((buf->uds_cookie & MAXRDCOOKIE) << RDOFFBITS) | (buf->
							      uds_offset &
							      DIREOF);
	return tmp;
}

/* Directory Branch read optimisation */
int raif_rule_readdir(dentry_t* dentry);


#endif /* __KERNEL__ */

#endif	/* not __RAIF_H_ */

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
